@@ -11,7 +11,7 @@
 # Author   : Kenneth J. Pronovici <pronovic@ieee.org>
 # Language : Python (>= 2.3)
 # Project  : Cedar Backup, release 2
-# Revision : $Id: __init__.py 695 2005-03-10 06:12:47Z pronovic $
+# Revision : $Id: __init__.py 1098 2007-01-31 05:14:25Z pronovic $
 # Purpose  : Provides package initialization
 #
 # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
@@ -21,15 +21,23 @@
 ########################################################################
 
 """
-Implements secure remote and local backups to CD-R or CD-RW.
+Implements local and remote backups to CD or DVD media.
 
-Cedar Backup is a Python package that supports secure backups of files
-on local and remote hosts to CD-R or CD-RW media.  The package is
-focused around weekly backups to a single disc, with the expectation
-that the disc will be changed or overwritten at the beginning of each
-week.  If your hardware is new enough, the script can write multisession
-discs, allowing you to add to a disc in a daily fashion.  Directories
-are backed up using tar and may be compressed using gzip or bzip2.
+Cedar Backup is a software package designed to manage system backups for a pool
+of local and remote machines.  Cedar Backup understands how to back up
+filesystem data as well as MySQL and PostgreSQL databases and Subversion
+repositories.  It can also be easily extended to support other kinds of data
+sources.
+
+Cedar Backup is focused around weekly backups to a single CD or DVD disc, with
+the expectation that the disc will be changed or overwritten at the beginning
+of each week.  If your hardware is new enough, Cedar Backup can write
+multisession discs, allowing you to add incremental data to a disc on a daily
+basis.
+
+Besides offering command-line utilities to manage the backup process, Cedar
+Backup provides a well-organized library of backup-related functionality,
+written in the Python programming language.
 
 @author: Kenneth J. Pronovici <pronovic@ieee.org>
 """
@@ -43,6 +51,6 @@ are backed up using tar and may be compressed using gzip or bzip2.
 # in the __all__ variable.
 
 __all__ = [ 'release', 'cli', 'actions', 'util', 'knapsack', 
-            'filesystem', 'peer', 'image', 'writer', 'config', 
-            'extend', ] 
+            'filesystem', 'peer', 'config', 
+            'actions', 'extend', 'writers', ] 
 
