@@ -8,7 +8,7 @@
 #
 # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
 #
-# Copyright (c) 2004-2005 Kenneth J. Pronovici.
+# Copyright (c) 2004-2007 Kenneth J. Pronovici.
 # All rights reserved.
 #
 # This program is free software; you can redistribute it and/or
@@ -27,11 +27,10 @@
 # Author   : Kenneth J. Pronovici <pronovic@ieee.org>
 # Language : Python (>= 2.3)
 # Project  : Cedar Backup, release 2
-# Revision : $Id: cli.py 749 2005-04-17 02:40:14Z pronovic $
+# Revision : $Id: cli.py 1076 2007-01-16 00:47:17Z pronovic $
 # Purpose  : Provides command-line interface implementation.
 #
 # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
-# This file was created with a width of 132 characters, and NO tabs.
 
 ########################################################################
 # Module documentation
@@ -88,9 +87,15 @@ import getopt
 
 # Cedar Backup modules
 from CedarBackup2.release import AUTHOR, EMAIL, VERSION, DATE, COPYRIGHT
-from CedarBackup2.util import RestrictedContentList, splitCommandLine, getFunctionReference, getUidGid, encodePath
+from CedarBackup2.util import RestrictedContentList, DirectedGraph, PathResolverSingleton
+from CedarBackup2.util import sortDict, splitCommandLine, executeCommand, getFunctionReference, getUidGid, encodePath
 from CedarBackup2.config import Config
-from CedarBackup2.action import executeCollect, executeStage, executeStore, executePurge, executeRebuild, executeValidate
+from CedarBackup2.actions.collect import executeCollect
+from CedarBackup2.actions.stage import executeStage
+from CedarBackup2.actions.store import executeStore
+from CedarBackup2.actions.purge import executePurge
+from CedarBackup2.actions.rebuild import executeRebuild
+from CedarBackup2.actions.validate import executeValidate
 
 
 ########################################################################
@@ -110,6 +115,8 @@ DEFAULT_LOGFILE    = "/var/log/cback.log"
 DEFAULT_OWNERSHIP  = [ "root", "adm", ]
 DEFAULT_MODE       = 0640
 
+REBUILD_INDEX      = 0        # can't run with anything else, anyway
+VALIDATE_INDEX     = 0        # can't run with anything else, anyway
 COLLECT_INDEX      = 100
 STAGE_INDEX        = 200
 STORE_INDEX        = 300
@@ -142,14 +149,13 @@ def cli():
    indicated actions.
 
    As a general rule, only the actions indicated on the command line will be
-   executed.   We will accept any of the standard actions and any of the
+   executed.   We will accept any of the built-in actions and any of the
    configured extended actions (which makes action list verification a two-
    step process).
 
-   The C{'all'} action has a special meaning.  Normally, it means that the
-   standard set of actions (collect, stage, store, purge) will all be executed,
-   in that order.  Extended actions will be ignored as part of the C{'all'}
-   action.
+   The C{'all'} action has a special meaning: it means that the built-in set of
+   actions (collect, stage, store, purge) will all be executed, in that order.
+   Extended actions will be ignored as part of the C{'all'} action.
 
    Raised exceptions always result in an immediate return.  Otherwise, we
    generally return when all specified actions have been completed.  Actions
@@ -218,12 +224,10 @@ def cli():
    try:
       logger.info("Configuration path is [%s]" % configPath)
       config = Config(xmlPath=configPath)
-      if config.extensions is not None:
-         actionSet = _ActionSet(options.actions, config.extensions.actions)
-      else:
-         actionSet = _ActionSet(options.actions, None)
+      _setupPathResolver(config)
+      actionSet = _ActionSet(options.actions, config.extensions, config.options.hooks)
    except Exception, e:
-      logger.error("Error reading configuration: %s" % e)
+      logger.error("Error reading or handling configuration: %s" % e)
       logger.info("Cedar Backup run completed with status 4.")
       return 4
 
@@ -258,22 +262,17 @@ class _ActionItem(object):
    """
    Class representing a single action to be executed.
 
-   This class represents a single action to be executed, and understands how to
-   execute those actions.   
+   This class represents a single named action to be executed, and understands
+   how to execute that action.   
 
-   Standard actions ("built-in" actions like collect, stage, etc.) are
-   instantiated in terms of a direct function reference, i.e.  to
-   C{executeCollect}.  Extended actions are instantiated in terms of an
-   C{ExtendedAction} object taken from configuration.
+   The built-in actions will use only the options and config values.  We also
+   pass in the config path so that extension modules can re-parse configuration
+   if they want to, to add in extra information.
 
-   The two different actions are executed directly.  In the case of standard
-   actions, the function reference is called directly.  In the case of extended
-   actions, a function reference is first derived (using
-   L{getFunctionReference}) and then called.
-
-   The standard actions will generally use the options and config values.  We
-   also pass in the config path so that extensions modules can re-parse
-   configuration if they want to, to add in extra information.
+   This class is also where pre-action and post-action hooks are executed.  An
+   action item is instantiated in terms of optional pre- and post-action hook
+   objects (config.ActionHook), which are then executed at the appropriate time
+   (if set).
 
    @note: The comparison operators for this class have been implemented to only
    compare based on the index value, and ignore all other values.  This is so
@@ -282,16 +281,24 @@ class _ActionItem(object):
    @sort: __init__, index, module, function 
    """
 
-   def __init__(self, index, function=None, extension=None):
+   def __init__(self, index, name, preHook, postHook, function):
       """
       Default constructor.
+
+      It's OK to pass C{None} for C{index}, C{preHook} or C{postHook}, but not
+      for C{name}.
+
       @param index: Index of the item (or C{None}).
+      @param name: Name of the action that is being executed.
+      @param preHook: Pre-action hook in terms of an C{ActionHook} object, or C{None}.
+      @param postHook: Post-action hook in terms of an C{ActionHook} object, or C{None}.
       @param function: Reference to function associated with item.
-      @param extension: C{ExtendedAction} object associated with item.
       """
       self.index = index
+      self.name = name
+      self.preHook = preHook
+      self.postHook = postHook
       self.function = function
-      self.extension = extension
 
    def __cmp__(self, other):
       """
@@ -311,7 +318,7 @@ class _ActionItem(object):
 
    def executeAction(self, configPath, options, config):
       """
-      Executes the action associated with an item.
+      Executes the action associated with an item, including hooks.
 
       See class notes for more details on how the action is executed.
 
@@ -321,11 +328,33 @@ class _ActionItem(object):
 
       @raise Exception: If there is a problem executing the action.
       """
-      if self.function is not None:
-         self.function(configPath, options, config)
-      else:
-         function = getFunctionReference(self.extension.module, self.extension.function)
-         function(configPath, options, config)
+      logger.debug("Executing [%s] action." % self.name)
+      if self.preHook is not None:
+         self._executeHook("pre-action", self.preHook)
+      self._executeAction(configPath, options, config)
+      if self.postHook is not None:
+         self._executeHook("post-action", self.postHook)
+      
+   def _executeAction(self, configPath, options, config):
+      """
+      Executes the action, specifically the function associated with the action.
+      @param configPath: Path to configuration file on disk.
+      @param options: Command-line options to be passed to action.
+      @param config: Parsed configuration to be passed to action.
+      """
+      name = "%s.%s" % (self.function.__module__, self.function.__name__)
+      logger.debug("Calling action function [%s], execution index [%d]" % (name, self.index))
+      self.function(configPath, options, config)
+
+   def _executeHook(self, type, hook):
+      """
+      Executes a hook command via L{util.executeCommand()}.
+      @param type: String describing the type of hook, for logging.
+      @param hook: Hook, in terms of a C{ActionHook} object.
+      """
+      logger.debug("Executing %s hook for action [%s]." % (type, hook.action))
+      fields = splitCommandLine(hook.command)
+      executeCommand(command=fields[0:1], args=fields[1:])
 
 
 ###################
@@ -337,40 +366,245 @@ class _ActionSet(object):
    """
    Class representing a set of actions to be executed.
 
-   This class does two different things.  First, it ensures that the actions
+   This class does three different things.  First, it ensures that the actions
    specified on the command-line are sensible.  The command-line can only list
-   either standard actions or extended actions specified in configuration.
+   either built-in actions or extended actions specified in configuration.
    Also, certain actions (in L{NONCOMBINE_ACTIONS}) cannot be combined with
    other actions.  
 
-   Second, it enforces an execution order on the specified actions.  Any time
-   actions are combined on the command line (either standard actions or
+   Second, the class enforces an execution order on the specified actions.  Any
+   time actions are combined on the command line (either built-in actions or
    extended actions), we must make sure they get executed in a sensible order.
 
+   Third, the class ensures that any pre-action or post-action hooks are
+   scheduled and executed appropriately.  Hooks are configured by building a
+   dictionary mapping between hook action name and command.  Pre-action hooks
+   are executed immediately before their associated action, and post-action
+   hooks are executed immediately after their associated action.
+
    @sort: __init__, executeActions
    """
 
-   def __init__(self, actions, extensions):
+   def __init__(self, actions, extensions, hooks):
       """
       Constructor for the C{_ActionSet} class.
 
+      This is kind of ugly, because the constructor has to set up a lot of data
+      before being able to do anything useful.  The following data structures
+      are initialized based on the input:
+   
+         - C{extensionNames}: List of extensions available in configuration
+         - C{preHookMap}: Mapping from action name to pre C{ActionHook}
+         - C{preHookMap}: Mapping from action name to post C{ActionHook}
+         - C{functionMap}: Mapping from action name to Python function
+         - C{indexMap}: Mapping from action name to execution index
+         - C{actionMap}: Mapping from action name to C{_ActionItem}
+
+      Once these data structures are set up, the command line is validated to
+      make sure only valid actions have been requested, and in a sensible
+      combination.  Then, all of the data is used to build C{self.actionSet},
+      the set of C{_ActionItem} object to be executed by C{executeActions()}.
+
       @param actions: Names of actions specified on the command-line.
-      @param extensions: List of extended actions (i.e. config.extensions.actions)
+      @param extensions: Extended action configuration (i.e. config.extensions)
+      @param hooks: List of pre- and post-action hooks (i.e. config.options.hooks)
 
       @raise ValueError: If one of the specified actions is invalid.
+      @raise ValueError: 
       """
-      extensionDict = {}
-      if extensions is not None:
-         for extension in extensions:
-            extensionDict[extension.name] = extension
-      _ActionSet._validateActions(actions, extensionDict.keys())
-      self.actionSet = _ActionSet._buildActionSet(actions, extensionDict)
+      extensionNames = _ActionSet._deriveExtensionNames(extensions)
+      (preHookMap, postHookMap) = _ActionSet._buildHookMaps(hooks)
+      functionMap = _ActionSet._buildFunctionMap(extensions)
+      indexMap = _ActionSet._buildIndexMap(extensions)
+      actionMap = _ActionSet._buildActionMap(extensionNames, functionMap, indexMap, preHookMap, postHookMap)
+      _ActionSet._validateActions(actions, extensionNames)
+      self.actionSet = _ActionSet._buildActionSet(actions, actionMap)
+
+   def _deriveExtensionNames(extensions):
+      """
+      Builds a list of extended actions that are available in configuration.
+      @param extensions: Extended action configuration (i.e. config.extensions)
+      @return: List of extended action names.
+      """
+      extensionNames = []
+      if extensions is not None and extensions.actions is not None:
+         for action in extensions.actions:
+            extensionNames.append(action.name)
+      return extensionNames
+   _deriveExtensionNames = staticmethod(_deriveExtensionNames)
+
+   def _buildHookMaps(hooks):
+      """
+      Build two mappings from action name to configured C{ActionHook}.
+      @param hooks: List of pre- and post-action hooks (i.e. config.options.hooks)
+      @return: Tuple of (pre hook dictionary, post hook dictionary).
+      """
+      preHookMap = {}
+      postHookMap = {}
+      if hooks is not None:
+         for hook in hooks:
+            if hook.before:
+               preHookMap[hook.action] = hook
+            elif hook.after:
+               postHookMap[hook.action] = hook
+      return (preHookMap, postHookMap)
+   _buildHookMaps = staticmethod(_buildHookMaps)
+
+   def _buildFunctionMap(extensions):
+      """
+      Builds a mapping from named action to action function.
+      @param extensions: Extended action configuration (i.e. config.extensions)
+      @return: Dictionary mapping action to function.
+      """
+      functionMap = {}
+      functionMap['rebuild'] = executeRebuild
+      functionMap['validate'] = executeValidate
+      functionMap['collect'] = executeCollect
+      functionMap['stage'] = executeStage
+      functionMap['store'] = executeStore
+      functionMap['purge'] = executePurge
+      if extensions is not None and extensions.actions is not None:
+         for action in extensions.actions:
+            functionMap[action.name] = getFunctionReference(action.module, action.function)
+      return functionMap
+   _buildFunctionMap = staticmethod(_buildFunctionMap)
+
+   def _buildIndexMap(extensions):
+      """
+      Builds a mapping from action name to proper execution index.
+
+      If extensions configuration is C{None}, or there are no configured
+      extended actions, the ordering dictionary will only include the built-in
+      actions and their standard indices.
+
+      Otherwise, if the extensions order mode is C{None} or C{"index"}, actions
+      will scheduled by explicit index; and if the extensions order mode is
+      C{"dependency"}, actions will be scheduled using a dependency graph.
+
+      @param extensions: Extended action configuration (i.e. config.extensions)
+
+      @return: Dictionary mapping action name to integer execution index.
+      """
+      indexMap = {}
+      if extensions is None or extensions.actions is None or extensions.actions == []:
+         logger.info("Action ordering will use 'index' order mode.")
+         indexMap['rebuild'] = REBUILD_INDEX;
+         indexMap['validate'] = VALIDATE_INDEX;
+         indexMap['collect'] = COLLECT_INDEX;
+         indexMap['stage'] = STAGE_INDEX;
+         indexMap['store'] = STORE_INDEX;
+         indexMap['purge'] = PURGE_INDEX;
+         logger.debug("Completed filling in action indices for built-in actions.")
+         logger.info("Action order will be: %s" % sortDict(indexMap))
+      else:
+         if extensions.orderMode is None or extensions.orderMode == "index":
+            logger.info("Action ordering will use 'index' order mode.")
+            indexMap['rebuild'] = REBUILD_INDEX;
+            indexMap['validate'] = VALIDATE_INDEX;
+            indexMap['collect'] = COLLECT_INDEX;
+            indexMap['stage'] = STAGE_INDEX;
+            indexMap['store'] = STORE_INDEX;
+            indexMap['purge'] = PURGE_INDEX;
+            logger.debug("Completed filling in action indices for built-in actions.")
+            for action in extensions.actions:
+               indexMap[action.name] = action.index
+            logger.debug("Completed filling in action indices for extended actions.")
+            logger.info("Action order will be: %s" % sortDict(indexMap))
+         else:
+            logger.info("Action ordering will use 'dependency' order mode.")
+            graph = DirectedGraph("dependencies")
+            graph.createVertex("rebuild")
+            graph.createVertex("validate")
+            graph.createVertex("collect")
+            graph.createVertex("stage")
+            graph.createVertex("store")
+            graph.createVertex("purge")
+            for action in extensions.actions:
+               graph.createVertex(action.name)
+            graph.createEdge("collect", "stage")   # Collect must run before stage, store or purge
+            graph.createEdge("collect", "store")
+            graph.createEdge("collect", "purge")
+            graph.createEdge("stage", "store")     # Stage must run before store or purge
+            graph.createEdge("stage", "purge")
+            graph.createEdge("store", "purge")     # Store must run before purge
+            for action in extensions.actions:
+               if action.dependencies.beforeList is not None:
+                  for vertex in action.dependencies.beforeList:
+                     try: 
+                        graph.createEdge(action.name, vertex)   # actions that this action must be run before
+                     except ValueError:
+                        logger.error("Dependency [%s] on extension [%s] is unknown." % (vertex, action.name))
+                        raise ValueError("Unable to determine proper action order due to invalid dependency.")
+               if action.dependencies.afterList is not None:
+                  for vertex in action.dependencies.afterList:
+                     try: 
+                        graph.createEdge(vertex, action.name)   # actions that this action must be run after
+                     except ValueError:
+                        logger.error("Dependency [%s] on extension [%s] is unknown." % (vertex, action.name))
+                        raise ValueError("Unable to determine proper action order due to invalid dependency.")
+            try:
+               ordering = graph.topologicalSort()
+               indexMap = dict([(ordering[i], i+1) for i in range(0, len(ordering))])
+               logger.info("Action order will be: %s" % ordering)
+            except ValueError:
+               logger.error("Unable to determine proper action order due to dependency recursion.")
+               logger.error("Extensions configuration is invalid (check for loops).")
+               raise ValueError("Unable to determine proper action order due to dependency recursion.")
+      return indexMap
+   _buildIndexMap = staticmethod(_buildIndexMap)
+
+   def _buildActionMap(extensionNames, functionMap, indexMap, preHookMap, postHookMap):
+      """
+      Builds a mapping from action name to list of C{_ActionItem} objects.
+
+      In most cases, the mapping from action name to C{_ActionItem} is 1:1.  The exception
+      is the "all" action, which is a special case.  However, a list is returned in all
+      cases, just for consistency later.
+
+      Each C{_ActionItem} will be created with a proper function reference and
+      index value for execution ordering.
+
+      @param extensionNames: List of valid extended action names 
+      @param functionMap: Dictionary mapping action name to Python function
+      @param indexMap: Dictionary mapping action name to integer execution index
+      @param preHookMap: Dictionary mapping action name to pre hooks (if any) for the action
+      @param postHookMap: Dictionary mapping action name to post hooks (if any) for the action
+
+      @return: Dictionary mapping action name to list of C{_ActionItem} objects.
+      """
+      actionMap = {}
+      for name in extensionNames + VALID_ACTIONS:
+         if name != 'all': # do this one later
+            function = functionMap[name]
+            index = indexMap[name]
+            (preHook, postHook) = _ActionSet._deriveHooks(name, preHookMap, postHookMap)
+            actionMap[name] = [ _ActionItem(index, name, preHook, postHook, function), ]
+      actionMap['all'] = actionMap['collect'] + actionMap['stage'] + actionMap['store'] + actionMap['purge']
+      return actionMap
+   _buildActionMap = staticmethod(_buildActionMap)
+
+   def _deriveHooks(action, preHookDict, postHookDict):
+      """
+      Derive pre- and post-action hooks, if any, associated with named action.
+      @param action: Name of action to look up
+      @param preHookDict: Dictionary mapping pre-action hooks to action name
+      @param postHookDict: Dictionary mapping post-action hooks to action name
+      @return Tuple (preHook, postHook) per mapping, with None values if there is no hook.
+      """
+      preHook = None
+      postHook = None
+      if preHookDict.has_key(action):
+         preHook = preHookDict[action]
+      if postHookDict.has_key(action):
+         postHook = postHookDict[action]
+      return (preHook, postHook)
+   _deriveHooks = staticmethod(_deriveHooks)
 
    def _validateActions(actions, extensionNames):
       """
       Validate that the set of specified actions is sensible.
 
-      Any specified action must either be a standard action or must be among
+      Any specified action must either be a built-in action or must be among
       the extended actions defined in configuration.  The actions from within
       L{NONCOMBINE_ACTIONS} may not be combined with other actions.
 
@@ -379,8 +613,6 @@ class _ActionSet(object):
 
       @raise ValueError: If one or more configured actions are not valid.
       """
-      if extensionNames is None: 
-         extensionNames = []
       if actions is None or actions == []:
          raise ValueError("No actions specified.")
       for action in actions:
@@ -391,44 +623,24 @@ class _ActionSet(object):
             raise ValueError("Action [%s] may not be combined with other actions." % action)
    _validateActions = staticmethod(_validateActions)
 
-   def _buildActionSet(actions, extensionDict):
+   def _buildActionSet(actions, actionMap):
       """
       Build set of actions to be executed.
 
-      The set of actions is built in the proper order, so C{executeActions}
+      The set of actions is built in the proper order, so C{executeActions} can
       spin through the set without thinking about it.  Since we've already validated
       that the set of actions is sensible, we don't take any precautions here to
       make sure things are combined properly.  If the action is listed, it will
-      be "scheduled" for implementation.
-
-      Every item in the action set is an C{_ActionItem}.
+      be "scheduled" for execution.
 
       @param actions: Names of actions specified on the command-line.
-      @param extensionDict: Dictionary mapping extension name to C{ExtendedAction} object.
+      @param actionMap: Dictionary mapping action name to C{_ActionItem} object.
 
       @return: Set of action items in proper order.
       """
       actionSet = []
       for action in actions:
-         if extensionDict.has_key(action):
-            actionSet.append(_ActionItem(extensionDict[action].index, extension=extensionDict[action]))
-         elif action == 'collect':
-            actionSet.append(_ActionItem(COLLECT_INDEX, function=executeCollect))
-         elif action == 'stage':
-            actionSet.append(_ActionItem(STAGE_INDEX, function=executeStage))
-         elif action == 'store':
-            actionSet.append(_ActionItem(STORE_INDEX, function=executeStore))
-         elif action == 'purge':
-            actionSet.append(_ActionItem(PURGE_INDEX, function=executePurge))
-         elif action == 'rebuild':
-            actionSet.append(_ActionItem(None, function=executeRebuild))
-         elif action == 'validate':
-            actionSet.append(_ActionItem(None, function=executeValidate))
-         elif action == 'all':
-            actionSet.append(_ActionItem(COLLECT_INDEX, function=executeCollect))
-            actionSet.append(_ActionItem(STAGE_INDEX, function=executeStage))
-            actionSet.append(_ActionItem(STORE_INDEX, function=executeStore))
-            actionSet.append(_ActionItem(PURGE_INDEX, function=executePurge))
+         actionSet.extend(actionMap[action])
       actionSet.sort()  # sort the actions in order by index
       return actionSet
    _buildActionSet = staticmethod(_buildActionSet)
@@ -436,10 +648,11 @@ class _ActionSet(object):
    def executeActions(self, configPath, options, config):
       """
       Executes all actions and extended actions, in the proper order.
-   
-      The built-in actions will generally use the options and config values.
-      We also pass in the config path so that extensions modules can re-parse
-      configuration if they want to, to add in extra information.
+
+      Each action (whether built-in or extension) is executed in an identical
+      manner.  The built-in actions will use only the options and config
+      values.  We also pass in the config path so that extension modules can
+      re-parse configuration if they want to, to add in extra information.
 
       @param configPath: Path to configuration file on disk.
       @param options: Command-line options to be passed to action functions.
@@ -680,6 +893,31 @@ def _setupDiskOutputLogging(outputLogger, logfile, options):
    outputLogger.addHandler(handler)
 
 
+################################
+# _setupPathResolver() function
+################################
+
+def _setupPathResolver(config):
+   """
+   Set up the path resolver singleton based on configuration.
+
+   Cedar Backup's path resolver is implemented in terms of a singleton, the
+   L{PathResolverSingleton} class.  This function takes options configuration,
+   converts it into the dictionary form needed by the singleton, and then
+   initializes the singleton.  After that, any function that needs to resolve
+   the path of a command can use the singleton.
+
+   @param config: Configuration
+   @type config: L{Config} object
+   """
+   mapping = {}
+   if config.options.overrides is not None:
+      for override in config.options.overrides:
+         mapping[override.command] = override.absolutePath
+   singleton = PathResolverSingleton()
+   singleton.fill(mapping)
+
+
 #########################################################################
 # Options class definition
 ########################################################################
