@@ -8,7 +8,7 @@
 #
 # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
 #
-# Copyright (c) 2005 Kenneth J. Pronovici.
+# Copyright (c) 2005,2007 Kenneth J. Pronovici.
 # All rights reserved.
 #
 # This program is free software; you can redistribute it and/or
@@ -27,11 +27,10 @@
 # Author   : Kenneth J. Pronovici <pronovic@ieee.org>
 # Language : Python (>= 2.3)
 # Project  : Official Cedar Backup Extensions
-# Revision : $Id: subversion.py 676 2005-03-07 06:59:43Z pronovic $
+# Revision : $Id: subversion.py 1046 2007-01-05 01:18:32Z pronovic $
 # Purpose  : Provides an extension to back up Subversion repositories.
 #
 # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
-# This file was created with a width of 132 characters, and NO tabs.
 
 ########################################################################
 # Module documentation
@@ -41,24 +40,27 @@
 Provides an extension to back up Subversion databases.
 
 This is a Cedar Backup extension used to back up Subversion repositories via
-the Cedar Backup command line.  It requires a new configurations section
-<subversion> and is intended to be run either immediately before or immediately
-after the standard collect action.  Aside from its own configuration, it
-requires the options and collect configuration sections in the standard Cedar
-Backup configuration file.
+the Cedar Backup command line.  Each Subversion repository can be backed using
+the same collect modes allowed for filesystems in the standard Cedar Backup
+collect action: weekly, daily, incremental.  
+
+This extension requires a new configuration section <subversion> and is
+intended to be run either immediately before or immediately after the standard
+collect action.  Aside from its own configuration, it requires the options and
+collect configuration sections in the standard Cedar Backup configuration file.
 
 There are two different kinds of Subversion repositories at this writing: BDB
-(Berkeley Database) and fsfs (a more CVS-like "filesystem within a filesystem").
-I personally only use the BDB version, because when I started using Subversion,
-that's all there was.  This extension only backs up BDB repositories.  If I
-ever start using fsfs, I'll add another function to back up that kind of 
-repository as well.  (Feel free to contribute code if you would like to see
-fsfs supported.)
-
-Each repository can be backed using the same collect modes allowed for
-filesystems in the standard Cedar Backup collect action: weekly, daily,
-incremental.  
+(Berkeley Database) and FSFS (a "filesystem within a filesystem").  This
+extension backs up both kinds of repositories in the same way, using C{svnadmin
+dump} in an incremental mode.
 
+It turns out that FSFS repositories can also be backed up just like any
+other filesystem directory.  If you would rather do that, then use the normal
+collect action.  This is probably simpler, although it carries its own 
+advantages and disadvantages (plus you will have to be careful to exclude
+the working directories Subversion uses when building an update to commit).
+Check the Subversion documentation for more information.
+ 
 @author: Kenneth J. Pronovici <pronovic@ieee.org>
 """
 
@@ -73,21 +75,13 @@ import pickle
 from bz2 import BZ2File
 from gzip import GzipFile
 
-# XML-related modules
-from xml.dom.ext.reader import PyExpat
-from xml.xpath import Evaluate
-from xml.parsers.expat import ExpatError
-from xml.dom.minidom import Node
-from xml.dom.minidom import getDOMImplementation
-from xml.dom.minidom import parseString
-from xml.dom.ext import PrettyPrint
-
 # Cedar Backup modules
-from CedarBackup2.config import addContainerNode, addStringNode
-from CedarBackup2.config import readChildren, readFirstChild, readString
+from CedarBackup2.xmlutil import createInputDom, addContainerNode, addStringNode
+from CedarBackup2.xmlutil import isElement, readChildren, readFirstChild, readString
 from CedarBackup2.config import VALID_COLLECT_MODES, VALID_COMPRESS_MODES
-from CedarBackup2.action import isStartOfWeek, buildNormalizedPath
-from CedarBackup2.util import executeCommand, ObjectTypeList, encodePath, changeOwnership
+from CedarBackup2.util import isStartOfWeek, buildNormalizedPath
+from CedarBackup2.util import resolveCommand, executeCommand
+from CedarBackup2.util import ObjectTypeList, encodePath, changeOwnership
 
 
 ########################################################################
@@ -114,26 +108,6 @@ class Repository(object):
    This is a base class used for validation later.  All subversion repository
    configuration (no matter how many types there are) must inherit from this
    class.
-   """
-
-   def __init__(self):
-      """
-      Constructor for the C{Repository} class.
-      """
-      raise NotImplementedError("Repository may not be instantiated.")
-
-
-########################################################################
-# BDBRepository class definition
-########################################################################
-
-class BDBRepository(Repository):
-
-   """
-   Class representing Subversion BDB (Berkeley Database) repository configuration.
-
-   The Subversion configuration information is used for backing up single
-   Subversion repository in BDB form.
 
    The following restrictions exist on data in this class:
 
@@ -146,13 +120,15 @@ class BDBRepository(Repository):
 
    def __init__(self, repositoryPath=None, collectMode=None, compressMode=None):
       """
-      Constructor for the C{BDBRepository} class.
+      Constructor for the C{Repository} class.
+
+      You should never directly instantiate this class.
       
       @param repositoryPath: Absolute path to a Subversion repository on disk.
       @param collectMode: Overridden collect mode for this directory.
       @param compressMode: Overridden compression mode for this directory.
       """
-      self._repositoryType = "BDB"
+      self._repositoryType = None
       self._repositoryPath = None
       self._collectMode = None
       self._compressMode = None
@@ -164,7 +140,7 @@ class BDBRepository(Repository):
       """
       Official string representation for class instance.
       """
-      return "BDBRepository(%s, %s, %s)" % (self.repositoryPath, self.collectMode, self.compressMode)
+      return "Repository(%s, %s, %s)" % (self.repositoryPath, self.collectMode, self.compressMode)
 
    def __str__(self):
       """
@@ -205,7 +181,6 @@ class BDBRepository(Repository):
    def _getRepositoryType(self):
       """
       Property target used to get the repository type.
-      For this class, this value is always "BDB".
       """
       return self._repositoryType
 
@@ -262,12 +237,90 @@ class BDBRepository(Repository):
       """
       return self._compressMode
 
-   repositoryType = property(_getRepositoryType, None, None, doc="Type of this repository (BDB).")
+   repositoryType = property(_getRepositoryType, None, None, doc="Type of this repository.")
    repositoryPath = property(_getRepositoryPath, _setRepositoryPath, None, doc="Path to the repository to collect.")
    collectMode = property(_getCollectMode, _setCollectMode, None, doc="Overridden collect mode for this repository.")
    compressMode = property(_getCompressMode, _setCompressMode, None, doc="Overridden compress mode for this repository.")
 
 
+########################################################################
+# BDBRepository class definition
+########################################################################
+
+class BDBRepository(Repository):
+
+   """
+   Class representing Subversion BDB (Berkeley Database) repository configuration.
+
+   The Subversion configuration information is used for backing up single
+   Subversion repository in BDB form.
+
+   The following restrictions exist on data in this class:
+
+      - The respository path must be absolute.
+      - The collect mode must be one of the values in L{VALID_COLLECT_MODES}.
+      - The compress mode must be one of the values in L{VALID_COMPRESS_MODES}.
+
+   @sort: __init__, __repr__, __str__, __cmp__, repositoryPath, collectMode, compressMode
+   """
+
+   def __init__(self, repositoryPath=None, collectMode=None, compressMode=None):
+      """
+      Constructor for the C{BDBRepository} class.
+      
+      @param repositoryPath: Absolute path to a Subversion repository on disk.
+      @param collectMode: Overridden collect mode for this directory.
+      @param compressMode: Overridden compression mode for this directory.
+      """
+      super(BDBRepository, self).__init__(repositoryPath, collectMode, compressMode)
+      self._repositoryType = "BDB"
+
+   def __repr__(self):
+      """
+      Official string representation for class instance.
+      """
+      return "BDBRepository(%s, %s, %s)" % (self.repositoryPath, self.collectMode, self.compressMode)
+
+
+########################################################################
+# FSFSRepository class definition
+########################################################################
+
+class FSFSRepository(Repository):
+
+   """
+   Class representing Subversion FSFS repository configuration.
+
+   The Subversion configuration information is used for backing up single
+   Subversion repository in FSFS form.
+
+   The following restrictions exist on data in this class:
+
+      - The respository path must be absolute.
+      - The collect mode must be one of the values in L{VALID_COLLECT_MODES}.
+      - The compress mode must be one of the values in L{VALID_COMPRESS_MODES}.
+
+   @sort: __init__, __repr__, __str__, __cmp__, repositoryPath, collectMode, compressMode
+   """
+
+   def __init__(self, repositoryPath=None, collectMode=None, compressMode=None):
+      """
+      Constructor for the C{FSFSRepository} class.
+      
+      @param repositoryPath: Absolute path to a Subversion repository on disk.
+      @param collectMode: Overridden collect mode for this directory.
+      @param compressMode: Overridden compression mode for this directory.
+      """
+      super(FSFSRepository, self).__init__(repositoryPath, collectMode, compressMode)
+      self._repositoryType = "FSFS"
+
+   def __repr__(self):
+      """
+      Official string representation for class instance.
+      """
+      return "FSFSRepository(%s, %s, %s)" % (self.repositoryPath, self.collectMode, self.compressMode)
+
+
 ########################################################################
 # SubversionConfig class definition
 ########################################################################
@@ -537,19 +590,20 @@ class LocalConfig(object):
       mode and compress mode are both optional, but the list of repositories
       must contain at least one entry.
 
-      Each BDB repository must contain a repository path, and then must be
-      either able to take collect mode and compress mode configuration from the
-      parent C{SubversionConfig} object, or must set each value on its own.  We
-      don't look at any other kinds of repositories that might be in the list.
+      Each BDB or FSFS repository must contain a repository path, and then must
+      be either able to take collect mode and compress mode configuration from
+      the parent C{SubversionConfig} object, or must set each value on its own.
+      We don't look at any other kinds of repositories that might be in the
+      list.
 
       @raise ValueError: If one of the validations fails.
       """
       if self.subversion is None:
          raise ValueError("Subversion section is required.")
       if self.subversion.repositories is None or len(self.subversion.repositories) < 1:
-         raise ValueError("At least one repository must be configured.")
+         raise ValueError("At least one Subversion repository must be configured.")
       for repository in self.subversion.repositories:
-         if repository.repositoryType == "BDB":
+         if repository.repositoryType in [ "BDB", "FSFS", ]:
             if repository.repositoryPath is None:
                raise ValueError("Each repository directory must set a repository path.")
             if self.subversion.collectMode is None and repository.collectMode is None:
@@ -597,12 +651,8 @@ class LocalConfig(object):
 
       @raise ValueError: If the XML cannot be successfully parsed.
       """
-      try:
-         xmlDom = PyExpat.Reader().fromString(xmlData)
-         parent = readFirstChild(xmlDom, "cb_config")
-         self._subversion = LocalConfig._parseSubversion(parent)
-      except (IOError, ExpatError), e:
-         raise ValueError("Unable to parse XML document: %s" % e)
+      (xmlDom, parentNode) = createInputDom(xmlData)
+      self._subversion = LocalConfig._parseSubversion(parentNode)
 
    def _parseSubversion(parent):
       """
@@ -648,8 +698,8 @@ class LocalConfig(object):
 
       The type field is optional, and if it isn't there we assume a BDB
       repositories.  Note that we will currently ignore any repository listing
-      which has a set type other than C{"BDB"}.  However, we will log a warning
-      if we do this.
+      which has a set type other than C{["BDB", "FSFS", ]}.  However, we will
+      log a warning if we do this.
 
       @param parent: Parent node to search beneath.
 
@@ -658,7 +708,7 @@ class LocalConfig(object):
       """
       lst = []
       for entry in readChildren(parent, "repository"):
-         if entry.nodeType == Node.ELEMENT_NODE:
+         if isElement(entry):
             repositoryType = readString(entry, "type")
             if repositoryType in [ None, "BDB", ]:    # BDB is the default type 
                repository = BDBRepository()
@@ -666,6 +716,12 @@ class LocalConfig(object):
                repository.collectMode = readString(entry, "collect_mode")
                repository.compressMode = readString(entry, "compress_mode")
                lst.append(repository)
+            elif repositoryType == "FSFS":
+               repository = FSFSRepository()
+               repository.repositoryPath = readString(entry, "abs_path")
+               repository.collectMode = readString(entry, "collect_mode")
+               repository.compressMode = readString(entry, "compress_mode")
+               lst.append(repository)
             else:
                logger.warn("Warning: Ignoring Subversion repository with unknown type [%s]." % repositoryType)
       if lst == []:
@@ -736,20 +792,21 @@ def executeAction(configPath, options, config):
    logger.debug("Full backup flag is [%s]" % fullBackup)
    if local.subversion.repositories is not None:
       for repository in local.subversion.repositories:
-         logger.debug("Working with repository [%s]." % repository.repositoryPath)
-         if repository.repositoryType == "BDB":
+         logger.debug("Working with repository [%s]" % repository.repositoryPath)
+         logger.debug("Repository type is [%s]" % repository.repositoryType)
+         if repository.repositoryType in ["BDB", "FSFS", ]:
             collectMode = _getCollectMode(local, repository)
             compressMode = _getCompressMode(local, repository)
             revisionPath = _getRevisionPath(config, repository)
-            backupPath = _getBackupPath(config, repository, compressMode)
             if fullBackup or (collectMode in ['daily', 'incr', ]) or (collectMode == 'weekly' and todayIsStart):
                logger.debug("Repository meets criteria to be backed up today.")
-               _backupBDBRepository(config, repository.repositoryPath, backupPath, 
-                                    revisionPath, fullBackup, collectMode, compressMode)
+               _backupRepository(config, repository.repositoryType,
+                                 repository.repositoryPath, revisionPath, 
+                                 fullBackup, collectMode, compressMode)
             else:
                logger.debug("Repository will not be backed up, per collect mode.")
          else:
-            logger.debug("Repository will not be backed up, since it is not type BDB.")
+            logger.debug("Repository will not be backed up, since it is not type BDB or FSFS.")
          logger.info("Completed backing up Subversion repository [%s]." % repository.repositoryPath)
    logger.info("Executed the Subversion extended action successfully.")
 
@@ -795,15 +852,18 @@ def _getRevisionPath(config, repository):
    logger.debug("Revision file path is [%s]" % revisionPath)
    return revisionPath
 
-def _getBackupPath(config, repository, compressMode):
+def _getBackupPath(config, repositoryPath, compressMode, startRevision, endRevision):
    """
    Gets the backup file path (including correct extension) associated with a repository.
    @param config: Config object.
-   @param repository: BDBRepository object.
+   @param repositoryPath: Path to the indicated repository
    @param compressMode: Compress mode to use for this repository.
+   @param startRevision: Starting repository revision.
+   @param endRevision: Ending repository revision.
    @return: Absolute path to the backup file associated with the repository.
    """
-   filename = "svndump-%s.txt" % buildNormalizedPath(repository.repositoryPath)
+   normalizedPath = buildNormalizedPath(repositoryPath)
+   filename = "svndump-%d:%d-%s.txt" % (startRevision, endRevision, normalizedPath)
    if compressMode == 'gzip':
       filename = "%s.gz" % filename
    elif compressMode == 'bzip2':
@@ -812,16 +872,16 @@ def _getBackupPath(config, repository, compressMode):
    logger.debug("Backup file path is [%s]" % backupPath)
    return backupPath
 
-def _backupBDBRepository(config, repositoryPath, backupPath, revisionPath, fullBackup, collectMode, compressMode):
+def _backupRepository(config, repositoryType, repositoryPath, revisionPath, fullBackup, collectMode, compressMode):
    """
-   Backs up an individual Subversion BDB repository.
+   Backs up an individual Subversion repository (either BDB or FSFS).
 
-   This internal method wraps the public method and adds some functionality
+   This internal method wraps the public methods and adds some functionality
    to work better with the extended action itself.
 
    @param config: Cedar Backup configuration.
+   @param repositoryType: Type of the repository (assumed to be BDB or FSFS).
    @param repositoryPath: Path to Subversion repository to back up.
-   @param backupPath: Path to backup file that will be written.
    @param revisionPath: Path used to store incremental revision information.
    @param fullBackup: Indicates whether this should be a full backup.
    @param collectMode: Collect mode to use.
@@ -845,9 +905,13 @@ def _backupBDBRepository(config, repositoryPath, backupPath, revisionPath, fullB
             logger.info("No need to back up repository [%s]; no new revisions." % repositoryPath)
             return
       logger.debug("Using incremental backup, revision: (%d, %d)." % (startRevision, endRevision))
+   backupPath = _getBackupPath(config, repositoryPath, compressMode, startRevision, endRevision)
    outputFile = _getOutputFile(backupPath, compressMode)
    try:
-      backupBDBRepository(repositoryPath, outputFile, startRevision, endRevision)
+      if repositoryType == "BDB":
+         backupBDBRepository(repositoryPath, outputFile, startRevision, endRevision)
+      else:
+         backupFSFSRepository(repositoryPath, outputFile, startRevision, endRevision)
    finally:
       outputFile.close()
    if not os.path.exists(backupPath):
@@ -959,7 +1023,7 @@ def backupBDBRepository(repositoryPath, backupFile, startRevision=None, endRevis
    @type endRevision: Integer value >= 0.
 
    @raise ValueError: If some value is missing or invalid.
-   @raise IOError: If there is a problem executing the MySQL dump.
+   @raise IOError: If there is a problem executing the Subversion dump.
    """
    if startRevision is None:
       startRevision = 0
@@ -972,9 +1036,69 @@ def backupBDBRepository(repositoryPath, backupFile, startRevision=None, endRevis
    if startRevision > endRevision:
       raise ValueError("Start revision must be <= end revision.")
    args = [ "dump", "--quiet", "-r%s:%s" % (startRevision, endRevision), "--incremental", repositoryPath, ]
-   result = executeCommand(SVNADMIN_COMMAND, args, returnOutput=False, ignoreStderr=True, doNotLog=True, outputFile=backupFile)[0]
+   command = resolveCommand(SVNADMIN_COMMAND)
+   result = executeCommand(command, args, returnOutput=False, ignoreStderr=True, doNotLog=True, outputFile=backupFile)[0]
    if result != 0:
-      raise IOError("Error [%d] executing Subversion dump for repository [%s]." % (result, repositoryPath))
+      raise IOError("Error [%d] executing Subversion dump for BDB repository [%s]." % (result, repositoryPath))
+   logger.debug("Completed dumping subversion repository [%s]." % repositoryPath)
+
+
+##################################
+# backupFSFSRepository() function
+##################################
+
+def backupFSFSRepository(repositoryPath, backupFile, startRevision=None, endRevision=None):
+   """
+   Backs up an individual Subversion FSFS repository.
+
+   The starting and ending revision values control an incremental backup.  If
+   the starting revision is not passed in, then revision zero (the start of the
+   repository) is assumed.  If the ending revision is not passed in, then the
+   youngest revision in the database will be used as the endpoint.
+
+   The backup data will be written into the passed-in back file.  Normally,
+   this would be an object as returned from C{open}, but it is possible to use
+   something like a C{GzipFile} to write compressed output.  The caller is
+   responsible for closing the passed-in backup file.
+
+   You should only use this function to back up an FSFS repository if you want
+   to use C{svnadmin dump} for your backup.  Use a normal filesystem copy
+   otherwise.  That will be simpler, and possibly less prone to problems like
+   updates to the repository in the middle of a backup.
+
+   @note: This function should either be run as root or as the owner of the
+   Subversion repository.
+
+   @param repositoryPath: Path to Subversion repository to back up
+   @type repositoryPath: String path representing Subversion FSFS repository on disk.
+
+   @param backupFile: Python file object to use for writing backup.
+   @type backupFile: Python file object as from C{open()} or C{file()}.
+   
+   @param startRevision: Starting repository revision to back up (for incremental backups)
+   @type startRevision: Integer value >= 0.
+
+   @param endRevision: Ending repository revision to back up (for incremental backups)
+   @type endRevision: Integer value >= 0.
+
+   @raise ValueError: If some value is missing or invalid.
+   @raise IOError: If there is a problem executing the Subversion dump.
+   """
+   if startRevision is None:
+      startRevision = 0
+   if endRevision is None:
+      endRevision = getYoungestRevision(repositoryPath)
+   if int(startRevision) < 0:
+      raise ValueError("Start revision must be >= 0.")
+   if int(endRevision) < 0:
+      raise ValueError("End revision must be >= 0.")
+   if startRevision > endRevision:
+      raise ValueError("Start revision must be <= end revision.")
+   args = [ "dump", "--quiet", "-r%s:%s" % (startRevision, endRevision), "--incremental", repositoryPath, ]
+   command = resolveCommand(SVNADMIN_COMMAND)
+   result = executeCommand(command, args, returnOutput=False, ignoreStderr=True, doNotLog=True, outputFile=backupFile)[0]
+   if result != 0:
+      raise IOError("Error [%d] executing Subversion dump for FSFS repository [%s]." % (result, repositoryPath))
    logger.debug("Completed dumping subversion repository [%s]." % repositoryPath)
 
 
@@ -998,7 +1122,8 @@ def getYoungestRevision(repositoryPath):
    @raise IOError: If there is a problem executing the C{svnlook} command.
    """
    args = [ 'youngest', repositoryPath, ]
-   (result, output) = executeCommand(SVNLOOK_COMMAND, args, returnOutput=True, ignoreStderr=True)
+   command = resolveCommand(SVNLOOK_COMMAND)
+   (result, output) = executeCommand(command, args, returnOutput=True, ignoreStderr=True)
    if result != 0:
       raise IOError("Error [%d] executing 'svnlook youngest' for repository [%s]." % (result, repositoryPath))
    if len(output) != 1:
