@@ -8,7 +8,7 @@
 #
 # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
 #
-# Copyright (c) 2004-2005 Kenneth J. Pronovici.
+# Copyright (c) 2004-2007 Kenneth J. Pronovici.
 # All rights reserved.
 #
 # This program is free software; you can redistribute it and/or
@@ -27,11 +27,10 @@
 # Author   : Kenneth J. Pronovici <pronovic@ieee.org>
 # Language : Python (>= 2.3)
 # Project  : Cedar Backup, release 2
-# Revision : $Id: filesystem.py 750 2005-04-18 19:25:23Z pronovic $
+# Revision : $Id: filesystem.py 1030 2007-01-03 04:59:12Z pronovic $
 # Purpose  : Provides filesystem-related objects.
 #
 # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
-# This file was created with a width of 132 characters, and NO tabs.
 
 ########################################################################
 # Module documentation
@@ -85,10 +84,10 @@ class FilesystemList(list):
    can add individual files or directories to the list, or can recursively add
    the contents of a directory.  The class also allows for up-front exclusions
    in several forms (all files, all directories, all items matching a pattern,
-   or all directories containing a specific "ignore file").  Symbolic links are
-   typically backed up non-recursively, i.e. the link to a directory is backed
-   up, but not the contents of that link (we don't want to deal with recursive
-   loops, etc.).
+   all items whose basename matches a pattern, or all directories containing a
+   specific "ignore file").  Symbolic links are typically backed up
+   non-recursively, i.e. the link to a directory is backed up, but not the
+   contents of that link (we don't want to deal with recursive loops, etc.).
 
    The custom methods such as L{addFile} will only add items if they exist on
    the filesystem and do not match any exclusions that are already in place.
@@ -106,12 +105,17 @@ class FilesystemList(list):
 
    @note: Regular expression patterns that apply to paths are assumed to be
    bounded at front and back by the beginning and end of the string, i.e. they
-   are treated as if they begin with C{^} and end with C{$}.
+   are treated as if they begin with C{^} and end with C{$}.  This is true
+   whether we are matching a complete path or a basename.
+
+   @note: Some platforms, like Windows, do not support soft links.  On those
+   platforms, the ignore-soft-links flag can be set, but it won't do any good
+   because the operating system never reports a file as a soft link.
 
    @sort: __init__, addFile, addDir, addDirContents, removeFiles, removeDirs,
           removeLinks, removeMatch, removeInvalid, normalize, validate, 
           excludeFiles, excludeDirs, excludeLinks, excludePaths, 
-          excludePatterns, ignoreFile 
+          excludePatterns, excludeBasenamePatterns, ignoreFile 
    """
 
 
@@ -127,12 +131,14 @@ class FilesystemList(list):
       self._excludeLinks = False
       self._excludePaths = None
       self._excludePatterns = None
+      self._excludeBasenamePatterns = None
       self._ignoreFile = None
       self.excludeFiles = False
       self.excludeLinks = False
       self.excludeDirs = False
       self.excludePaths = []
       self.excludePatterns = []
+      self.excludeBasenamePatterns = []
       self.ignoreFile = None
 
 
@@ -220,6 +226,21 @@ class FilesystemList(list):
       """
       return self._excludePatterns
 
+   def _setExcludeBasenamePatterns(self, value):
+      """
+      Property target used to set the exclude basename patterns list.
+      A C{None} value is converted to an empty list.
+      """
+      self._excludeBasenamePatterns = []
+      if value is not None:
+         self._excludeBasenamePatterns.extend(value)
+
+   def _getExcludeBasenamePatterns(self):
+      """
+      Property target used to get the exclude basename patterns list.
+      """
+      return self._excludeBasenamePatterns
+
    def _setIgnoreFile(self, value):
       """
       Property target used to set the ignore file.
@@ -241,7 +262,10 @@ class FilesystemList(list):
    excludeDirs = property(_getExcludeDirs, _setExcludeDirs, None, "Boolean indicating whether directories should be excluded.")
    excludeLinks = property(_getExcludeLinks, _setExcludeLinks, None, "Boolean indicating whether soft links should be excluded.")
    excludePaths = property(_getExcludePaths, _setExcludePaths, None, "List of absolute paths to be excluded.")
-   excludePatterns = property(_getExcludePatterns, _setExcludePatterns, None, "List of regular expression patterns to be excluded.")
+   excludePatterns = property(_getExcludePatterns, _setExcludePatterns, None, 
+                              "List of regular expression patterns (matching complete path) to be excluded.")
+   excludeBasenamePatterns = property(_getExcludeBasenamePatterns, _setExcludeBasenamePatterns, 
+                                      None, "List of regular expression patterns (matching basename) to be excluded.")
    ignoreFile = property(_getIgnoreFile, _setIgnoreFile, None, "Name of file which will cause directory contents to be ignored.")
    
 
@@ -281,7 +305,12 @@ class FilesystemList(list):
          if re.compile(r"^%s$" % pattern).match(path):
             logger.debug("Path [%s] is excluded based on pattern [%s]." % (path, pattern))
             return 0
+      for pattern in self.excludeBasenamePatterns:
+         if re.compile(r"^%s$" % pattern).match(os.path.basename(path)):
+            logger.debug("Path [%s] is excluded based on basename pattern [%s]." % (path, pattern))
+            return 0
       self.append(path)
+      logger.debug("Added file to list: [%s]" % path)
       return 1
 
    def addDir(self, path):
@@ -319,17 +348,23 @@ class FilesystemList(list):
          if re.compile(r"^%s$" % pattern).match(path):
             logger.debug("Path [%s] is excluded based on pattern [%s]." % (path, pattern))
             return 0
+      for pattern in self.excludeBasenamePatterns:
+         if re.compile(r"^%s$" % pattern).match(os.path.basename(path)):
+            logger.debug("Path [%s] is excluded based on basename pattern [%s]." % (path, pattern))
+            return 0
       self.append(path)
+      logger.debug("Added directory to list: [%s]" % path)
       return 1
 
-   def addDirContents(self, path):
+   def addDirContents(self, path, recursive=True, addSelf=True):
       """
       Adds the contents of a directory to the list.
 
       The path must exist and must be a directory or a link to a directory.
       The contents of the directory (as well as the directory path itself) will
       be recursively added to the list, subject to any exclusions that are in
-      place.  
+      place.  If you only want the directory and its immediate contents to be
+      added, then pass in C{recursive=False}.
 
       @note: If a directory's absolute path matches an exclude pattern or path,
       or if the directory contains the configured ignore file, then the
@@ -341,17 +376,19 @@ class FilesystemList(list):
       only be added by name, not recursively.   Any invalid soft links (i.e.
       soft links that point to non-existent items) will be silently ignored.
 
-      @note: The L{excludeDirs} flag only controls whether any given soft link
-      path itself is added to the list once it has been discovered.  It does
-      I{not} modify any behavior related to directory recursion.
-   
       @note: The L{excludeDirs} flag only controls whether any given directory
       path itself is added to the list once it has been discovered.  It does
       I{not} modify any behavior related to directory recursion.
-
+   
       @param path: Directory path whose contents should be added to the list
       @type path: String representing a path on disk
 
+      @param recursive: Indicates whether directory contents should be added recursively.
+      @type recursive: Boolean value
+
+      @param addSelf: Indicates whether the directory itself should be added to the list.
+      @type addSelf: Boolean value
+
       @return: Number of items recursively added to the list
 
       @raise ValueError: If path is not a directory or does not exist.
@@ -359,9 +396,9 @@ class FilesystemList(list):
       """
       path = encodePath(path)
       path = normalizeDir(path)
-      return self._addDirContentsRecursive(path)
+      return self._addDirContentsInternal(path, recursive=recursive, includePath=addSelf)
 
-   def _addDirContentsRecursive(self, path, includePath=True):
+   def _addDirContentsInternal(self, path, includePath=True, recursive=True):
       """
       Internal implementation of C{addDirContents}.
 
@@ -375,6 +412,7 @@ class FilesystemList(list):
 
       @param path: Directory path whose contents should be added to the list.
       @param includePath: Indicates whether to include the path as well as contents.
+      @param recursive: Indicates whether directory contents should be added recursively.
 
       @return: Number of items recursively added to the list
 
@@ -391,6 +429,10 @@ class FilesystemList(list):
          if re.compile(r"^%s$" % pattern).match(path):
             logger.debug("Path [%s] is excluded based on pattern [%s]." % (path, pattern))
             return added
+      for pattern in self.excludeBasenamePatterns:
+         if re.compile(r"^%s$" % pattern).match(os.path.basename(path)):
+            logger.debug("Path [%s] is excluded based on basename pattern [%s]." % (path, pattern))
+            return added
       if self.ignoreFile is not None and os.path.exists(os.path.join(path, self.ignoreFile)):
          logger.debug("Path [%s] is excluded based on ignore file." % path)
          return added
@@ -404,7 +446,10 @@ class FilesystemList(list):
             if os.path.islink(entrypath):
                added += self.addDir(entrypath)
             else:
-               added += self._addDirContentsRecursive(entrypath)
+               if recursive:
+                  added += self._addDirContentsInternal(entrypath)
+               else:
+                  added += self.addDir(entrypath)
       return added
 
 
@@ -531,7 +576,12 @@ class FilesystemList(list):
       is faster to call this method than to call the L{removeFiles},
       L{removeDirs} or L{removeLinks} methods individually.  If you know which
       patterns you will want to remove ahead of time, you may be better off
-      setting L{excludePatterns} before adding items to the list.
+      setting L{excludePatterns} or L{excludeBasenamePatterns} before adding
+      items to the list.
+
+      @note: Unlike when using the exclude lists, the pattern here is I{not}
+      bounded at the front and the back of the string.  You can use any pattern
+      you want.
 
       @param pattern: Regular expression pattern representing entries to remove
 
@@ -747,8 +797,8 @@ class BackupFileList(FilesystemList):
       Python hashing algorithms.
 
       In my tests using a 110 MB file on CD, the original implementation
-      requires 111 seconds while this implementation requires only 40-45
-      seconds, which is a pretty substantial speed-up.  
+      requires 111 seconds.  This implementation requires only 40-45 seconds,
+      which is a pretty substantial speed-up.  
 
       Practice shows that reading in around 4kB (4096 bytes) at a time yields
       the best performance.  Smaller reads are quite a bit slower, and larger
@@ -764,14 +814,16 @@ class BackupFileList(FilesystemList):
       @raise OSError: If the file cannot be opened.
       """
       s = sha.new()
-      f = open(path)
+      f = open(path, mode="rb")  # in case platform cares about binary reads
       readBytes = 4096  # see notes above
       while(readBytes > 0):
          readString = f.read(readBytes)
          s.update(readString)
          readBytes = len(readString)
       f.close()
-      return s.hexdigest()
+      digest = s.hexdigest()
+      logger.debug("Generated digest [%s] for file [%s]." % (digest, path))
+      return digest
    _generateDigest = staticmethod(_generateDigest)
 
    def generateFitted(self, capacity, algorithm="worst_fit"):
@@ -809,7 +861,7 @@ class BackupFileList(FilesystemList):
       elif algorithm == "alternate_fit": return alternateFit(table, capacity)[0]
       else: raise ValueError("Algorithm [%s] is invalid." % algorithm);
 
-   def generateTarfile(self, path, mode='tar', ignore=False):
+   def generateTarfile(self, path, mode='tar', ignore=False, flat=False):
       """
       Creates a tar file containing the files in the list.
 
@@ -822,6 +874,10 @@ class BackupFileList(FilesystemList):
       the extra functionality out-weighs the disadvantage of not being
       "standard".
 
+      If you pass in C{flat=True}, then a "flat" archive will be created, and
+      all of the files will be added to the root of the archive.  So, the file
+      C{/tmp/something/whatever.txt} would be added as just C{whatever.txt}.
+
       By default, the whole method call fails if there are problems adding any
       of the files to the archive, resulting in an exception.  Under these
       circumstances, callers are advised that they might want to call
@@ -858,6 +914,9 @@ class BackupFileList(FilesystemList):
       @param ignore: Indicates whether to ignore certain errors.
       @type ignore: Boolean
 
+      @param flat: Creates "flat" archive by putting all items in root 
+      @type flat: Boolean
+
       @raise ValueError: If mode is not valid
       @raise ValueError: If list is empty
       @raise ValueError: If the path could not be encoded properly.
@@ -874,23 +933,35 @@ class BackupFileList(FilesystemList):
          tar.posix = False    # make a GNU-compatible archive without file length limits
          for entry in self:
             try:
-               tar.add(entry, recursive=False)
-            except tarfile.TarError:
+               if flat:
+                  tar.add(entry, arcname=os.path.basename(entry), recursive=False)
+               else:
+                  tar.add(entry, recursive=False)
+            except tarfile.TarError, e:
                if not ignore:
-                  raise
+                  raise e
                logger.info("Unable to add file [%s]; going on anyway." % entry)
             except OSError, e:
                if not ignore:
                   raise tarfile.TarError(e)
                logger.info("Unable to add file [%s]; going on anyway." % entry)
          tar.close()
-      except tarfile.TarError:
+      except tarfile.ReadError, e:
+         try: tar.close()
+         except: pass
          if os.path.exists(path): 
             try: os.remove(path) 
             except: pass
-         raise
+         raise tarfile.ReadError("Unable to open [%s]; maybe directory doesn't exist?" % path)
+      except tarfile.TarError, e:
+         try: tar.close()
+         except: pass
+         if os.path.exists(path): 
+            try: os.remove(path) 
+            except: pass
+         raise e
 
-   def removeUnchanged(self, digestMap):
+   def removeUnchanged(self, digestMap, captureDigest=False):
       """
       Removes unchanged entries from the list.
    
@@ -901,40 +972,77 @@ class BackupFileList(FilesystemList):
 
       This method offers a convenient way for callers to filter unneeded
       entries from a list.  The idea is that a caller will capture a digest map
-      from C{generateDigestMap} immediately before beginning a backup, and will
-      save off that map using C{pickle} or some other method.  Then, the caller
-      could use this method sometime in the future to filter out any unchanged
-      files based on the saved-off map.
+      from C{generateDigestMap} at some point in time (perhaps the beginning of
+      the week), and will save off that map using C{pickle} or some other
+      method.  Then, the caller could use this method sometime in the future to
+      filter out any unchanged files based on the saved-off map.
 
-      For performance reasons, this method actually ends up rebuilding the list
-      from scratch.  First, we build a temporary dictionary containing all of
-      the items from the original list.  Then, we remove items as needed from
-      the dictionary (which is faster than the equivalent operation on a list).
-      Finally, we replace the contents of the current list based on the keys
-      left in the dictionary.  
+      If C{captureDigest} is passed-in as C{True}, then digest information will
+      be captured for the entire list before the removal step occurs using the
+      same rules as in L{generateDigestMap}.  The check will involve a lookup
+      into the complete digest map.  
 
-      We only generate a digest value for files we actually need to check, and
-      we'll ignore any entry in the list which isn't a file that currently
-      exists on disk.
+      If C{captureDigest} is passed in as C{False}, we will only generate a
+      digest value for files we actually need to check, and we'll ignore any
+      entry in the list which isn't a file that currently exists on disk.
+
+      The return value varies depending on C{captureDigest}, as well.  To
+      preserve backwards compatibility, if C{captureDigest} is C{False}, then
+      we'll just return a single value representing the number of entries
+      removed.  Otherwise, we'll return a tuple of C{(entries removed, digest
+      map)}.  The returned digest map will be in exactly the form returned by
+      L{generateDigestMap}.
+
+      @note: For performance reasons, this method actually ends up rebuilding
+      the list from scratch.  First, we build a temporary dictionary containing
+      all of the items from the original list.  Then, we remove items as needed
+      from the dictionary (which is faster than the equivalent operation on a
+      list).  Finally, we replace the contents of the current list based on the
+      keys left in the dictionary.  This should be transparent to the caller.
 
       @param digestMap: Dictionary mapping file name to digest value.
       @type digestMap: Map as returned from L{generateDigestMap}.
 
+      @param captureDigest: Indicates that digest information should be captured.
+      @type captureDigest: Boolean
+
       @return: Number of entries removed
       """
-      removed = 0
-      table = {}
-      for entry in self:
-         table[entry] = None
-      for entry in digestMap.keys():
-         if table.has_key(entry):
+      if captureDigest:
+         removed = 0
+         table = {}
+         captured = {}
+         for entry in self:
             if os.path.isfile(entry) and not os.path.islink(entry):
-               digest = BackupFileList._generateDigest(entry)
-               if digest == digestMap[entry]:
-                  removed += 1
-                  del table[entry]
-      self[:] = table.keys()
-      return removed
+               table[entry] = BackupFileList._generateDigest(entry)
+               captured[entry] = table[entry]
+            else:
+               table[entry] = None
+         for entry in digestMap.keys():
+            if table.has_key(entry):
+               if table[entry] is not None:  # equivalent to file/link check in other case
+                  digest = table[entry]
+                  if digest == digestMap[entry]:
+                     removed += 1
+                     del table[entry]
+                     logger.debug("Discarded unchanged file [%s]." % entry)
+         self[:] = table.keys()
+         return (removed, captured)
+      else:
+         removed = 0
+         table = {}
+         for entry in self:
+            table[entry] = None
+         for entry in digestMap.keys():
+            if table.has_key(entry):
+               if os.path.isfile(entry) and not os.path.islink(entry):
+                  digest = BackupFileList._generateDigest(entry)
+                  if digest == digestMap[entry]:
+                     removed += 1
+                     del table[entry]
+                     logger.debug("Discarded unchanged file [%s]." % entry)
+         self[:] = table.keys()
+         return removed
 
 
 ########################################################################
@@ -975,14 +1083,15 @@ class PurgeItemList(FilesystemList):
    # Add methods
    ##############
 
-   def addDirContents(self, path):
+   def addDirContents(self, path, recursive=True, addSelf=False):
       """
       Adds the contents of a directory to the list.
 
       The path must exist and must be a directory or a link to a directory.
       The contents of the directory (but I{not} the directory path itself) will
       be recursively added to the list, subject to any exclusions that are in
-      place.  
+      place.  If you only want the directory and its contents to be added, then
+      pass in C{recursive=False}.
 
       @note: If a directory's absolute path matches an exclude pattern or path,
       or if the directory contains the configured ignore file, then the
@@ -1005,6 +1114,11 @@ class PurgeItemList(FilesystemList):
       @param path: Directory path whose contents should be added to the list
       @type path: String representing a path on disk
 
+      @param recursive: Indicates whether directory contents should be added recursively.
+      @type recursive: Boolean value
+   
+      @param addSelf: Ignored in this subclass.
+
       @return: Number of items recursively added to the list
 
       @raise ValueError: If path is not a directory or does not exist.
@@ -1012,7 +1126,7 @@ class PurgeItemList(FilesystemList):
       """
       path = encodePath(path)
       path = normalizeDir(path)
-      return super(PurgeItemList, self)._addDirContentsRecursive(path, includePath=False)
+      return super(PurgeItemList, self)._addDirContentsInternal(path, includePath=False, recursive=recursive)
 
 
    ##################
@@ -1117,7 +1231,7 @@ def normalizeDir(path):
 
    @return: Normalized path, which should be equivalent to the original.
    """
-   if path[-1:] == os.sep:
+   if path != os.sep and path[-1:] == os.sep:
       return path[:-1]
    return path
 
