@@ -11,7 +11,7 @@
 # Author   : Kenneth J. Pronovici <pronovic@ieee.org>
 # Language : Python (>= 2.3)
 # Project  : Cedar Backup, release 2
-# Revision : $Id: release.py 756 2005-04-23 17:49:50Z pronovic $
+# Revision : $Id: release.py 1100 2007-01-31 05:16:03Z pronovic $
 # Purpose  : Provides location to maintain release information.
 #
 # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
@@ -33,8 +33,8 @@ Provides location to maintain version information.
 
 AUTHOR      = "Kenneth J. Pronovici"
 EMAIL       = "pronovic@ieee.org"
-COPYRIGHT   = "2004-2005"
-VERSION     = "2.4.2"
-DATE        = "23 Apr 2005"
+COPYRIGHT   = "2004-2007"
+VERSION     = "2.10.1"
+DATE        = "30 Jan 2007"
 URL         = "http://cedar-solutions.com/software/cedar-backup"
 
