@@ -9,7 +9,7 @@
 #
 # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
 #
-# Copyright (c) 2004-2005 Kenneth J. Pronovici.
+# Copyright (c) 2004-2007 Kenneth J. Pronovici.
 # All rights reserved.
 #
 # This program is free software; you can redistribute it and/or
@@ -28,20 +28,19 @@
 # Author   : Kenneth J. Pronovici <pronovic@ieee.org>
 # Language : Python (>= 2.3)
 # Project  : Cedar Backup, release 2
-# Revision : $Id: test.py 662 2005-03-07 04:47:51Z pronovic $
+# Revision : $Id: test.py 1090 2007-01-31 02:09:11Z pronovic $
 # Purpose  : Run all of the unit tests for the project.
 #
 # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
-# This file was created with a width of 132 characters, and NO tabs.
 
 ########################################################################
 # Notes
 ########################################################################
 
 """
-Run all of the CedarBackup2 unit tests.
+Run the CedarBackup2 unit tests.
 
-We run all of the unit tests here at once so we can get one big success or
+This script runs all of the unit tests at once so we can get one big success or
 failure result, rather than 20 different smaller results that we somehow have
 to aggregate together to get the "big picture".  This is done by creating and
 running one big unit test suite based on the suites in the individual unit test
@@ -63,15 +62,20 @@ want to make sure we are running the correct 'test' package - not one found
 elsewhere on the user's path - since 'test' could be a relatively common name
 for a package.
 
-Finally, this script might be used people who won't have an environment that
-allows running all of the tests, especially certain tests related to remote
-connectivity, loopback filesystems, etc.  Most people should run the script
-with no arguments.  This will result in a "reduced feature set" test suite with
-no surprising system, kernel or network dependencies.  People who understand
-what they're doing can put "full" as one of the arguments on the command-line,
-and they'll get all of the available tests (or they can dig further and
-explicitly set certain environment variables to get more precise control over
-what will be tested).
+Most people will want to run the script with no arguments.  This will result in
+a "reduced feature set" test suite that covers all of the available test
+suites, but executes only those tests with no surprising system, kernel or
+network dependencies.
+
+If "full" is specified as one of the command-line arguments, then all of the
+unit tests will be run, including those that require a specialized environment.
+For instance, some tests require remote connectivity, a loopback filesystem,
+etc.
+
+Other arguments on the command line are assumed to be named tests, so for
+instance passing "config" runs only the tests for config.py.  Any number of
+individual tests may be listed on the command line, and unknown values will
+simply be ignored.
 
 @note: Even if you run this test with the C{python2.3} interpreter, some of the
 individual unit tests require the C{python} interpreter.  In particular, the
@@ -99,28 +103,36 @@ def main():
 
    """
    Main routine for program.
+   @return: Integer 0 upon success, integer 1 upon failure.
    """
 
    # Check the Python version.  We require 2.3 or greater.
    try:
       if map(int, [sys.version_info[0], sys.version_info[1]]) < [2, 3]:
          print "Python version 2.3 or greater required, sorry."
-         sys.exit(1)
+         return 1
    except:
       # sys.version_info isn't available before 2.0
       print "Python version 2.3 or greater required, sorry."
-      sys.exit(1)
+      return 1
 
-   # Check for the correct CedarBackup2 location
-   if os.path.exists(os.path.join(".", "CedarBackup2", "filesystem.py")):
-      sys.path.insert(0, ".")
-   elif os.path.basename(os.getcwd()) == "test" and os.path.exists(os.path.join("..", "CedarBackup2", "filesystem.py")):
-      sys.path.insert(0, "..")
-   else:
-      print "WARNING: CedarBackup2 modules were not found in the expected"
-      print "location.  If the import succeeds, you may be using an"
-      print "unexpected version of CedarBackup2."
-      print ""
+   # Check for the correct CedarBackup2 location and import utilities
+   try:
+      if os.path.exists(os.path.join(".", "CedarBackup2", "filesystem.py")):
+         sys.path.insert(0, ".")
+      elif os.path.basename(os.getcwd()) == "test" and os.path.exists(os.path.join("..", "CedarBackup2", "filesystem.py")):
+         sys.path.insert(0, "..")
+      else:
+         print "WARNING: CedarBackup2 modules were not found in the expected"
+         print "location.  If the import succeeds, you may be using an"
+         print "unexpected version of CedarBackup2."
+         print ""
+      from CedarBackup2.util import nullDevice
+   except ImportError, e:
+      print "Failed to import CedarBackup2 util module: %s" % e
+      print "You must either run the unit tests from the CedarBackup2 source"
+      print "tree, or properly set the PYTHONPATH enviroment variable."
+      return 1
 
    # Import the unit test modules
    try:
@@ -137,52 +149,74 @@ def main():
       import test.knapsacktests as knapsacktests
       import test.filesystemtests as filesystemtests
       import test.peertests as peertests
-      import test.imagetests as imagetests
-      import test.writertests as writertests
+      import test.writersutiltests as writersutiltests
+      import test.cdwritertests as cdwritertests
+      import test.dvdwritertests as dvdwritertests
       import test.configtests as configtests
       import test.clitests as clitests
       import test.mysqltests as mysqltests
+      import test.postgresqltests as postgresqltests
       import test.subversiontests as subversiontests
+      import test.mboxtests as mboxtests
+      import test.encrypttests as encrypttests
    except ImportError, e:
       print "Failed to import CedarBackup2 unit test module: %s" % e
       print "You must either run the unit tests from the CedarBackup2 source"
       print "tree, or properly set the PYTHONPATH enviroment variable."
-      sys.exit(1)
-
-   # Set flags in the environment to control tests
-   if "full" in sys.argv:
-      os.environ["PEERTESTS_FULL"] = "Y"
-      os.environ["IMAGETESTS_FULL"] = "Y"
-   else:
-      os.environ["PEERTESTS_FULL"] = "N"
-      os.environ["IMAGETESTS_FULL"] = "N"
+      return 1
 
    # Set up logging to discard everything
-   handler = logging.FileHandler(filename="/dev/null")
+   devnull = nullDevice()
+   handler = logging.FileHandler(filename=devnull)
    handler.setLevel(logging.NOTSET)
    logger = logging.getLogger("CedarBackup2")
    logger.setLevel(logging.NOTSET)
    logger.addHandler(handler)
 
+   # Get a list of program arguments
+   args = sys.argv[1:]
+
+   # Set flags in the environment to control tests
+   if "full" in args:
+      os.environ["PEERTESTS_FULL"] = "Y"
+      os.environ["WRITERSUTILTESTS_FULL"] = "Y"
+      os.environ["ENCRYPTTESTS_FULL"] = "Y"
+      args.remove("full") # remainder of list will be specific tests to run, if any
+   else:
+      os.environ["PEERTESTS_FULL"] = "N"
+      os.environ["WRITERSUTILTESTS_FULL"] = "N"
+      os.environ["ENCRYPTTESTS_FULL"] = "N"
+
    # Print a starting banner
    print "\n*** Running CedarBackup2 unit tests."
 
+   # Make a list of tests to run
+   unittests = { }
+   if args == [] or "util" in args: unittests["util"] = utiltests.suite()
+   if args == [] or "knapsack" in args: unittests["knapsack"] = knapsacktests.suite()
+   if args == [] or "filesystem" in args: unittests["filesystem"] = filesystemtests.suite()
+   if args == [] or "peer" in args: unittests["peer"] = peertests.suite()
+   if args == [] or "writersutil" in args: unittests["writersutil"] = writersutiltests.suite()
+   if args == [] or "cdwriter" in args: unittests["cdwriter"] = cdwritertests.suite()
+   if args == [] or "dvdwriter" in args: unittests["dvdwriter"] = dvdwritertests.suite()
+   if args == [] or "config" in args: unittests["config"] = configtests.suite()
+   if args == [] or "cli" in args: unittests["cli"] = clitests.suite()
+   if args == [] or "mysql" in args: unittests["mysql"] = mysqltests.suite()
+   if args == [] or "postgresql" in args: unittests["postgresql"] = postgresqltests.suite()
+   if args == [] or "subversion" in args: unittests["subversion"] = subversiontests.suite()
+   if args == [] or "mbox" in args: unittests["mbox"] = mboxtests.suite()
+   if args == [] or "encrypt" in args: unittests["encrypt"] = encrypttests.suite()
+   if args != []: print "*** Executing specific tests: %s" % unittests.keys()
+
    # Create and run the test suite
    print ""
-   suite = unittest.TestSuite((
-                               utiltests.suite(),
-                               knapsacktests.suite(), 
-                               filesystemtests.suite(),
-                               peertests.suite(),
-                               imagetests.suite(),
-                               writertests.suite(),
-                               configtests.suite(),
-                               clitests.suite(),
-                               mysqltests.suite(),
-                               subversiontests.suite(),
-                              ))
-   unittest.TextTestRunner(verbosity=1).run(suite)
+   suite = unittest.TestSuite(unittests.values())
+   result = unittest.TextTestRunner(verbosity=1).run(suite)
    print ""
+   if not result.wasSuccessful():
+      return 1
+   else:
+      return 0
 
 
 ########################################################################
@@ -191,5 +225,6 @@ def main():
 
 # Run the main routine if the module is executed rather than sourced
 if __name__ == '__main__':
-   main()
+   result = main()
+   sys.exit(result)
 
